# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

"""
GraphCollector classes for IR
"""

from itertools import chain
from codetiming import Timer

try:
    from graphviz import Digraph, nohtml

    HAVE_IR_GRAPH = True
    """Indicate wheater the graphviz package is available."""
except ImportError:
    HAVE_IR_GRAPH = False

from loki.tools import JoinableStringList, is_iterable, as_tuple
from loki.ir.visitor import Visitor


__all__ = ["HAVE_IR_GRAPH", "GraphCollector", "ir_graph"]


class GraphCollector(Visitor):
    """
    Convert a given IR tree to a node and edge list via the visit mechanism.

    This serves as base class for backends and provides a number of helpful
    routines that ease implementing automatic recursion and line wrapping.
    It is adapted from the Stringifier in "pprint.py". It doubles as a means
    to produce a human readable graph representation of the IR, which is
    useful for debugging purposes and first visualization.

    Parameters
    ----------
    linewidth : int, optional
        The line width limit after which to break a line.
    symgen : str, optional
        A function handle that accepts a :any:`pymbolic.primitives.Expression`
        and produces a string representation for that.
    show_comments : bool, optional, default: False
        Whether to show comments in the output
    show_expressions : bool, optional, default: False
        Whether to further expand expressions in the output
    """

    def __init__(
        self, show_comments=False, show_expressions=False, linewidth=40, symgen=str
    ):
        super().__init__()
        self.linewidth = linewidth
        self._symgen = symgen
        self._id = 0
        self._id_map = {}
        self.show_comments = show_comments
        self.show_expressions = show_expressions

    @property
    def symgen(self):
        """
        Formatter for expressions.
        """
        return self._symgen

    def join_items(self, items, sep=", ", separable=True):
        """
        Concatenate a list of items into :any:`JoinableStringList`.

        The return value can be passed to :meth:`format_line` or
        :meth:`format_node` or converted to a string with `str`, using
        the :any:`JoinableStringList` as an argument.
        Upon expansion, lines will be wrapped automatically to stay within
        the linewidth limit.

        Parameters
        ----------
        items : list
            The list of strings to be joined.
        sep : str, optional
            The separator to be inserted between items.
        separable : bool, optional
            Allow line breaks between individual :data:`items`.

        Returns
        -------
        :any:`JoinableStringList`
        """
        return JoinableStringList(
            items,
            sep=sep,
            width=self.linewidth,
            cont="\n",
            separable=separable,
        )

    def format_node(self, name, *items):
        """
        Default format for a node.

        Creates a string of the form ``<name[, attribute, attribute, ...]>``.
        """
        content = ""
        if items:
            content = self.format_line("<", name, " ", self.join_items(items), ">")
        else:
            content = self.format_line("<", name, ">")

        # disregard all quotes to ensure nice graphviz label behaviour
        return content.replace('"', "")

    def format_line(self, *items, comment=None, no_wrap=False):
        """
        Format a line by concatenating all items and applying indentation while observing
        the allowed line width limit.

        Note that the provided comment will simply be extended to the line and no line
        width limit will be enforced for that.

        Parameters
        ----------
        items : list
            The items to be put on that line.
        comment : str
            An optional inline comment to be put at the end of the line.
        no_wrap : bool
            Disable line wrapping.

        Returns
        -------
        str the string of the current line, potentially including line breaks if
                 required to observe the line width limit.
        """

        if no_wrap:
            # Simply concatenate items and extend the comment
            line = "".join(str(item) for item in items)
        else:
            # Use join_items to concatenate items
            line = str(self.join_items(items, sep=""))
        if comment:
            return line + comment
        return line

    def visit_all(self, item, *args, **kwargs):
        """
        Convenience function to call :meth:`visit` for all given arguments.

        If only a single argument is given that is iterable,
        :meth:`visit` is called on all of its elements instead.
        """
        if is_iterable(item) and not args:
            return chain.from_iterable(
                as_tuple(self.visit(i, **kwargs) for i in item if i is not None)
            )
        return list(
            chain.from_iterable(
                as_tuple(
                    self.visit(i, **kwargs) for i in [item, *args] if i is not None
                )
            )
        )

    def __add_node(self, node, **kwargs):
        """
        Adds a node to the graphical representation of the IR. Utilizes the
        formatting provided by :meth:`format_node`.

        Parameters
        ----------
        node: :any: `Node` object
        kwargs["shape"]: str, optional (default: "oval")
        kwargs["label"]: str, optional (default: format_node(repr(node)))
        kwargs["parent"]: :any: `Node` object, optional (default: None)
            If not available no edge is drawn.

        Returns
        -------
        list[tuple[dict[str,str], dict[str,str]]]]
            A list of a tuple of a node and potentially a edge information
        """
        label = kwargs.get("label", "")

        if label == "":
            label = self.format_node(repr(node))

        try:
            live_symbols = "live: [" + ", ".join(
                str(symbol) for symbol in node.live_symbols
            )
            defines_symbols = "defines: [" + ", ".join(
                str(symbol) for symbol in node.defines_symbols
            )
            uses_symbols = "uses: [" + ", ".join(
                str(symbol) for symbol in node.uses_symbols
            )
            label = self.format_line(
                label,
                "\n",
                live_symbols,
                "], ",
                defines_symbols,
                "], ",
                uses_symbols,
                "]",
            )
        except (RuntimeError, KeyError, AttributeError) as _:
            pass

        shape = kwargs.get("shape", "oval")

        node_key = str(id(node))
        if node_key not in self._id_map:
            self._id_map[node_key] = str(self._id)
            self._id += 1

        node_info = {
            "name": str(self._id_map[node_key]),
            "label": nohtml(str(label)),
            "shape": str(shape),
        }

        parent = kwargs.get("parent")
        edge_info = {}
        if parent:
            parent_id = self._id_map[str(id(parent))]
            child_id = self._id_map[str(id(node))]
            edge_info = {"tail_name": str(parent_id), "head_name": str(child_id)}

        return [(node_info, edge_info)]

    # Handler for outer objects
    def visit_Module(self, o, **kwargs):
        """
        Add a :any:`Module`, mark parent node and visit all "spec" and "subroutine" nodes.

        Returns
        -------
        list[tuple[dict[str,str], dict[str,str]]]]
            An extended list of tuples of a node and potentially a edge information
        """
        node_edge_info = self.__add_node(o, **kwargs)
        kwargs["parent"] = o

        node_edge_info.extend(self.visit(o.spec, **kwargs))
        node_edge_info.extend(self.visit_all(o.contains, **kwargs))

        return node_edge_info

    def visit_Subroutine(self, o, **kwargs):
        """
        Add a :any:`Subroutine`, mark parent node and visit all "docstring", "spec", "body", "members" nodes.

        Returns
        -------
        list[tuple[dict[str,str], dict[str,str]]]]
            An extended list of tuples of a node and potentially a edge information
        """
        node_edge_info = self.__add_node(o, **kwargs)
        kwargs["parent"] = o

        node_edge_info.extend(self.visit(o.docstring, **kwargs))
        node_edge_info.extend(self.visit(o.spec, **kwargs))
        node_edge_info.extend(self.visit(o.body, **kwargs))
        node_edge_info.extend(self.visit_all(o.contains, **kwargs))

        return node_edge_info

    # Handler for AST base nodes
    def visit_Comment(self, o, **kwargs):
        """
        Enables turning off comments.

        Returns
        -------
        list[tuple[dict[str,str], dict[str,str]]]]
            An extended list of tuples of a node and potentially a edge information, or list of nothing.
        """
        if self.show_comments:
            return self.visit_Node(o, **kwargs)
        return []

    visit_CommentBlock = visit_Comment

    def visit_Node(self, o, **kwargs):
        """
        Add a :any:`Node`, mark parent and visit all children.

        Returns
        -------
        list[tuple[dict[str,str], dict[str,str]]]]
            An extended list of tuples of a node and potentially a edge information
        """
        node_edge_info = self.__add_node(o, **kwargs)
        kwargs["parent"] = o

        node_edge_info.extend(self.visit_all(o.children, **kwargs))
        return node_edge_info

    def visit_Expression(self, o, **kwargs):
        """
        Dispatch routine to add nodes utilizing expression tree stringifier,
        mark parent and stop.

        Returns
        -------
        list[tuple[dict[str,str], dict[str,str]]]]
            An extended list of tuples of a node and potentially a edge information or list of nothing.
        """
        if self.show_expressions:
            content = self.symgen(o)
            parent = kwargs.get("parent")
            return self.__add_node(o, label=content, parent=parent, shape="box")
        return []

    def visit_tuple(self, o, **kwargs):
        """
        Recurse for each item in the tuple.
        """
        return self.visit_all(o, **kwargs)

    visit_list = visit_tuple

    def visit_Conditional(self, o, **kwargs):
        """
        Add a :any:`Conditional`, mark parent and visit first body then else body.

        Returns
        -------
        list[tuple[dict[str,str], dict[str,str]]]]
            An extended list of tuples of a node and potentially a edge information
        """
        parent = kwargs.get("parent")
        label = self.symgen(o.condition)
        node_edge_info = self.__add_node(o, label=label, parent=parent, shape="diamond")
        kwargs["parent"] = o
        node_edge_info.extend(self.visit_all(o.body, **kwargs))

        if o.else_body:
            node_edge_info.extend(self.visit_all(o.else_body, **kwargs))
        return node_edge_info


def ir_graph(ir, show_comments=False, show_expressions=False, linewidth=40, symgen=str):
    """
    Pretty-print the given IR using :class:`GraphCollector`.

    Parameters
    ----------
    ir : :any:`Node`
        The IR node starting from which to produce the tree
    show_comments : bool, optional, default: False
        Whether to show comments in the output
    show_expressions : bool, optional, default: False
        Whether to further expand expressions in the output
    """

    if not HAVE_IR_GRAPH:
        raise ImportError("ir_graph is not available.")

    log = "[Loki::Graph Visualization] Created graph visualization in {:.2f}s"

    with Timer(text=log):
        graph_representation = GraphCollector(
            show_comments, show_expressions, linewidth, symgen
        )
        node_edge_info = [
            item for item in graph_representation.visit(ir) if item is not None
        ]

        graph = Digraph()
        graph.attr(rankdir="LR")
        for node_info, edge_info in node_edge_info:
            if node_info:
                graph.node(**node_info)
            if edge_info:
                graph.edge(**edge_info)
        return graph



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest

from loki import Sourcefile, Subroutine
from loki.expression import symbols as sym, parse_expr
from loki.frontend import available_frontends, OMNI
from loki.ir import (
    nodes as ir, FindNodes, FindVariables, FindTypedSymbols,
    SubstituteExpressions, SubstituteStringExpressions,
    FindLiterals, FindRealLiterals
)


@pytest.mark.parametrize('frontend', available_frontends())
def test_expression_finder_retrieval_function(frontend, tmp_path):
    """
    Verify that expression finder visitors work as intended and remain
    functional if re-used
    """
    fcode = """
module some_mod
    implicit none
contains
    function some_func() result(ret)
        integer :: ret
        ret = 1
    end function some_func

    subroutine other_routine
        integer :: var, tmp
        var = 5 + some_func()
    end subroutine other_routine
end module some_mod
    """.strip()

    source = Sourcefile.from_source(fcode, frontend=frontend, xmods=[tmp_path])

    expected_ts = {'var', 'some_func'}
    expected_vars = ('var',)

    # Instantiate the first expression finder and make sure it works as expected
    find_ts = FindTypedSymbols()
    assert find_ts.visit(source['other_routine'].body) == expected_ts

    # Verify that it works also on a repeated invocation
    assert find_ts.visit(source['other_routine'].body) == expected_ts

    # Instantiate the second expression finder and make sure it works as expected
    find_vars = FindVariables(unique=False)
    assert find_vars.visit(source['other_routine'].body) == expected_vars

    # Make sure the first expression finder still works
    assert find_ts.visit(source['other_routine'].body) == expected_ts


@pytest.mark.parametrize('frontend', available_frontends())
def test_find_variables(frontend, tmp_path):
    """ Test that :any:`FindVariables` finds all symbol uses. """

    fcode_external = """
module external_mod
implicit none
contains
subroutine rick(dave, never)
  real(kind=8), intent(inout) :: dave, never
end subroutine rick
end module external_mod
    """
    fcode = """
module test_mod
  use external_mod, only: rick
  implicit none

  type my_type
    real(kind=8) :: never
    real(kind=8), pointer :: give_you(:)
  end type my_type

contains

  subroutine test_routine(n, a, b, gonna)
    integer, intent(in) :: n
    real(kind=8), intent(inout) :: a, b(n)
    type(my_type), intent(inout) :: gonna
    integer :: i

    associate(will=>gonna%never, up=>n)
    do i=1, n
      b(i) = b(i) + a
    end do

    call rick(will, never=gonna%give_you(up))
    end associate
  end subroutine test_routine
end module test_mod
    """
    _ = Sourcefile.from_source(fcode_external, frontend=frontend, xmods=[tmp_path])
    source = Sourcefile.from_source(fcode, frontend=frontend, xmods=[tmp_path])
    routine = source['test_routine']

    # Test unique=True|False using the spec
    expected = ['n', 'a', 'gonna', 'i', 'b(n)']
    spec_vars = FindVariables(unique=True).visit(routine.spec)
    assert len(spec_vars) == 5
    assert all(v in spec_vars for v in expected)

    spec_vars = FindVariables(unique=False).visit(routine.spec)
    assert len(spec_vars) == 6
    assert all(v in spec_vars for v in expected)
    assert len([v for v in spec_vars if v == 'n']) == 2  # two occurences of 'n'

    # Test retrieval with associates and keyword arg calls
    expected = [
        'will', 'gonna', 'gonna%never', 'up', 'n', 'i', 'b(i)', 'a',
        'rick', 'gonna%give_you(up)'
    ]
    body_vars = FindVariables(unique=True).visit(routine.body)
    assert len(body_vars) == 10
    assert all(v in body_vars for v in expected)

@pytest.mark.parametrize('frontend', available_frontends())
def test_find_literals(frontend):
    """
    Test that :any:`FindLiterals` finds all literals
    and :any:`FindRealLiterals` all real/float literals.
    """
    fcode = """
subroutine test_find_literals()
  implicit none
  integer :: n, n1
  real(kind=8) :: x

  n = 1 + 5 + 42
  x = 1.0 / 10.5
  n1 = int(B'00000')
  if (.TRUE.) then
    call some_func(x, some_string='string_kwarg')
  endif

end subroutine test_find_literals
"""
    expected_int_literals = ('1', '5', '42')
    expected_real_literals = ('1.0', '10.5')
    # Omni evaluates BOZ constants, so it creates IntegerLiteral instead...
    expected_intrinsic_literals = ("B'00000'",) if frontend != OMNI else ('0',)
    expected_logic_literals = ('True',)
    expected_string_literals = ('string_kwarg',)
    expected_literals = expected_int_literals + expected_real_literals +\
            expected_intrinsic_literals + expected_logic_literals +\
            expected_string_literals
    routine = Subroutine.from_source(fcode, frontend=frontend)
    literals = FindLiterals().visit(routine.body)
    assert sorted(list(expected_literals)) == sorted([str(literal.value) for literal in literals])
    real_literals = FindRealLiterals().visit(routine.body)
    assert sorted(list(expected_real_literals)) == sorted([str(literal.value) for literal in real_literals])
    real_literals_isinstance = [literal for literal in literals if isinstance(literal, sym.FloatLiteral)]
    assert sorted(list(expected_real_literals)) == sorted([str(literal.value) for literal in real_literals_isinstance])


@pytest.mark.parametrize('frontend', available_frontends())
def test_substitute_expressions(frontend):
    """ Test symbol replacement with :any:`Expression` symbols. """

    fcode = """
subroutine test_routine(n, a, b)
  implicit none
  integer, intent(in) :: n
  real(kind=8), intent(inout) :: a, b(n)
  real(kind=8) :: c(n)
  integer :: i

  associate(d => a)
  do i=1, n
    c(i) = b(i) + a
  end do

  call another_routine(n, a, c(:), a2=d)

  end associate
end subroutine test_routine
"""
    routine = Subroutine.from_source(fcode, frontend=frontend)

    calls = FindNodes(ir.CallStatement).visit(routine.body)
    assoc = FindNodes(ir.Associate).visit(routine.body)[0]
    assert calls[0].arguments == ('n', 'a', 'c(:)')
    assert calls[0].kwarguments == (('a2', 'd'),)

    n = routine.variable_map['n']
    i = routine.variable_map['i']
    a = routine.variable_map['a']
    b_i = parse_expr('b(i)', scope=routine)
    c_r = parse_expr('c(:)', scope=routine)
    d = parse_expr('d', scope=assoc)
    expr_map = {
        n: sym.Sum((n, sym.Product((-1, sym.Literal(1))))),
        b_i: b_i.clone(dimensions=sym.Sum((i, sym.Literal(1)))),
        c_r: c_r.clone(dimensions=sym.Range((sym.Literal(1), sym.Literal(2)))),
        a: d,
        d: a,
    }
    routine.body = SubstituteExpressions(expr_map).visit(routine.body)

    loops = FindNodes(ir.Loop).visit(routine.body)
    assert loops[0].bounds == '1:n-1'
    assigns = FindNodes(ir.Assignment).visit(routine.body)
    assert assigns[0].lhs == 'c(i)' and assigns[0].rhs == 'b(i+1) + d'
    calls = FindNodes(ir.CallStatement).visit(routine.body)
    assert calls[0].arguments == ('n - 1', 'd', 'c(1:2)')
    assert calls[0].kwarguments == (('a2', 'a'),)


@pytest.mark.parametrize('frontend', available_frontends())
def test_substitute_string_expressions(frontend):
    """ Test symbol replacement with symbol string mappping. """

    fcode = """
subroutine test_routine(n, a, b)
  implicit none
  integer, intent(in) :: n
  real(kind=8), intent(inout) :: a, b(n)
  real(kind=8) :: c(n)
  integer :: i

  associate(d => a)
  do i=1, n
    c(i) = b(i) + a
  end do

  call another_routine(n, a, c(:), a2=d)

  end associate
end subroutine test_routine
"""
    routine = Subroutine.from_source(fcode, frontend=frontend)

    calls = FindNodes(ir.CallStatement).visit(routine.body)
    assoc = FindNodes(ir.Associate).visit(routine.body)[0]
    assert calls[0].arguments == ('n', 'a', 'c(:)')
    assert calls[0].kwarguments == (('a2', 'd'),)

    expr_map = {
        'n': 'n - 1',
        'b(i)': 'b(i+1)',
        'c(:)': 'c(1:2)',
        'a': 'd',
        'd': 'a',
    }
    # Note that we need to use the associate block here, as it defines 'd'
    routine.body = SubstituteStringExpressions(expr_map, scope=assoc).visit(routine.body)

    loops = FindNodes(ir.Loop).visit(routine.body)
    assert loops[0].bounds == '1:n-1'
    assigns = FindNodes(ir.Assignment).visit(routine.body)
    assert assigns[0].lhs == 'c(i)' and assigns[0].rhs == 'b(i+1) + d'
    calls = FindNodes(ir.CallStatement).visit(routine.body)
    assert calls[0].arguments == ('n - 1', 'd', 'c(1:2)')
    assert calls[0].kwarguments == (('a2', 'a'),)



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest

from loki import Module, Subroutine, as_tuple
from loki.frontend import available_frontends

from loki.transformations import PragmaModelTransformation
from loki.ir import FindNodes, Pragma

def check_pragma(pragma, keyword, content, check_for_equality=True):
    assert pragma.keyword == keyword
    if check_for_equality:
        assert pragma.content == content
    else:
        for _content in as_tuple(content):
            assert _content in pragma.content

@pytest.mark.parametrize('frontend', available_frontends())
@pytest.mark.parametrize('directive', [None, 'openacc', 'omp-gpu'])
@pytest.mark.parametrize('keep_loki_pragmas', [True, False])
def test_transform_pragma_model(tmp_path, frontend, directive, keep_loki_pragmas):
    """
    Test Pragma model trafo for different directives/flavors.
    """
    fcode_mod = """
    module some_mod
      integer :: a, b
      !$loki create device(a, b)
    end module some_mod
    """.strip()

    fcode = """
subroutine some_func(ret)
  implicit none
  integer, intent(out) :: ret
  integer :: tmp1, tmp2, tmp3, tmp4, jk

  !$loki create device(tmp1, tmp2)
  !$loki update device(tmp1) host(tmp2)
  !$loki unstructured-data in(tmp1, tmp2) create(tmp3, tmp4)
  !$loki end unstructured-data out(tmp2, tmp3, tmp4) delete(tmp1)
  !$loki structured-data in(tmp1) out(tmp2) inout(tmp3) create(tmp4)
  !$loki end structured-data in(tmp1) out(tmp2) inout(tmp3) create(tmp4)
  !$loki loop gang private(tmp1) vlength(128)
  !$loki end loop gang
  !$loki loop vector private(tmp2)
  !$loki end loop vector
  !$loki loop seq
  !$loki end loop seq
  !$loki routine vector
  !$loki routine seq
  !$loki device-present vars(tmp1, tmp2)
  !$loki end device-present vars(tmp1, tmp2)
  !$loki device-ptr vars(tmp1, tmp2)
  !$loki end device-ptr vars(tmp1, tmp2)
  !$loki unmapped-directive whatever(tmp1) foo(tmp2)
  ! misspelled by purpose
  !$loki create drvice(tmp1)

end subroutine some_func
    """.strip()

    routine = Subroutine.from_source(fcode, frontend=frontend, xmods=[tmp_path])
    module = Module.from_source(fcode_mod, frontend=frontend, xmods=[tmp_path])

    pragma_model_trafo = PragmaModelTransformation(directive=directive,
            keep_loki_pragmas=keep_loki_pragmas)
    pragma_model_trafo.transform_subroutine(routine)
    pragma_model_trafo.transform_module(module)

    # CHECK MODULE
    pragmas = FindNodes(Pragma).visit(module.spec)
    if directive == 'openacc':
        check_pragma(pragmas[0], 'acc', 'declare create(a, b)')
    if directive == 'omp-gpu':
        check_pragma(pragmas[0], 'omp', 'declare target(a, b)')
    if directive is None and keep_loki_pragmas:
        check_pragma(pragmas[0], 'loki', 'create device(a, b)')

    # CHECK ROUTINE
    pragmas = FindNodes(Pragma).visit(routine.ir)
    if directive == 'openacc':
        args = (('acc', 'declare create(tmp1, tmp2)'),
                ('acc', ('update', 'device(tmp1)', 'self(tmp2)'), False),
                ('acc', ('enter data', 'copyin(tmp1, tmp2)', 'create(tmp3, tmp4)'), False),
                ('acc', ('exit data', 'copyout(tmp2, tmp3, tmp4)', 'delete(tmp1)'), False),
                ('acc', ('data', 'copyin(tmp1)', 'copy(tmp3)', 'copyout(tmp2)', 'create(tmp4)'), False),
                ('acc', 'end data'),
                ('acc', ('parallel loop gang', 'private(tmp1)', 'vector_length(128)'), False),
                ('acc', 'end parallel loop'),
                ('acc', ('loop vector', 'private(tmp2)'), False),
                ('loki', 'end loop vector'),
                ('acc', 'loop seq'),
                ('loki', 'end loop seq'),
                ('acc', 'routine vector'),
                ('acc', 'routine seq'),
                ('acc', 'data present(tmp1, tmp2)'),
                ('acc', 'end data'),
                ('acc', 'data deviceptr(tmp1, tmp2)'),
                ('acc', 'end data'),
                ('loki', 'unmapped-directive whatever(tmp1) foo(tmp2)'),
                ('loki', 'create drvice(tmp1)'))
    if directive == 'omp-gpu':
        args = (('omp', 'declare target(tmp1, tmp2)'),
                ('omp', ('target update', 'to(tmp1)', 'from(tmp2)'), False),
                ('omp', ('target enter data', 'map(to: tmp1, tmp2)', 'map(alloc: tmp3, tmp4)'), False),
                ('omp', ('target exit data', 'map(from: tmp2, tmp3, tmp4)', 'map(delete: tmp1)'), False),
                ('omp', ('target data', 'map(to: tmp1)', 'map(tofrom: tmp3)',
                    'map(from: tmp2)', 'map(alloc: tmp4)'), False),
                ('omp', 'end target data'),
                ('omp', ('target teams distribute', 'thread_limit(128)'), False),
                ('omp', 'end target teams distribute'),
                ('omp', 'parallel do'),
                ('omp', 'end parallel do'),
                ('loki', 'loop seq'),
                ('loki', 'end loop seq'),
                ('loki', 'routine vector'),
                ('omp', 'declare target'),
                ('loki', ('device-present', 'vars(tmp1, tmp2)'), False),
                ('loki', ('end device-present', 'vars(tmp1, tmp2)'), False),
                ('loki', ('device-ptr', 'vars(tmp1, tmp2)'), False),
                ('loki', ('end device-ptr', 'vars(tmp1, tmp2)'), False),
                ('loki', 'unmapped-directive whatever(tmp1) foo(tmp2)'),
                ('loki', 'create drvice(tmp1)'))
    if directive is None:
        args = (('loki', 'create device(tmp1, tmp2)'),
                ('loki', ('update', 'device(tmp1)', 'host(tmp2)'), False),
                ('loki', ('unstructured-data', 'in(tmp1, tmp2)', 'create(tmp3, tmp4)'), False),
                ('loki', ('end unstructured-data', 'out(tmp2, tmp3, tmp4)', 'delete(tmp1)'), False),
                ('loki', ('structured-data', 'in(tmp1)', 'out(tmp2)', 'inout(tmp3)', 'create(tmp4)'), False),
                ('loki', ('end structured-data', 'in(tmp1)', 'out(tmp2)', 'inout(tmp3)', 'create(tmp4)'), False),
                ('loki', ('loop gang', 'private(tmp1)', 'vlength(128)'), False),
                ('loki', 'end loop gang'),
                ('loki', ('loop vector', 'private(tmp2)'), False),
                ('loki', 'end loop vector'),
                ('loki', 'loop seq'),
                ('loki', 'end loop seq'),
                ('loki', 'routine vector'),
                ('loki', 'routine seq'),
                ('loki', ('device-present', 'vars(tmp1, tmp2)'), False),
                ('loki', ('end device-present', 'vars(tmp1, tmp2)'), False),
                ('loki', ('device-ptr', 'vars(tmp1, tmp2)'), False),
                ('loki', ('end device-ptr', 'vars(tmp1, tmp2)'), False),
                ('loki', 'unmapped-directive whatever(tmp1) foo(tmp2)'),
                ('loki', 'create drvice(tmp1)'))

    if not keep_loki_pragmas:
        args = tuple(arg for arg in args if arg[0] != 'loki')

    for pragma, _args in zip(pragmas, args):
        check_pragma(pragma, *_args)



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from loki.expression import symbols as sym
from loki.frontend import (
    parse_omni_ast, parse_fparser_ast, get_fparser_node,
    parse_regex_source
)
from loki.ir import (
    nodes as ir, FindNodes, Transformer, ExpressionTransformer,
    pragmas_attached
)
from loki.logging import debug
from loki.program_unit import ProgramUnit
from loki.tools import as_tuple, CaseInsensitiveDict
from loki.types import BasicType, ProcedureType, SymbolAttributes



__all__ = ['Subroutine']


class Subroutine(ProgramUnit):
    """
    Class to handle and manipulate a single subroutine.

    Parameters
    ----------
    name : str
        Name of the subroutine.
    args : iterable of str, optional
        The names of the dummy args.
    docstring : tuple of :any:`Node`, optional
        The subroutine docstring in the original source.
    spec : :any:`Section`, optional
        The spec of the subroutine.
    body : :any:`Section`, optional
        The body of the subroutine.
    contains : :any:`Section`, optional
        The internal-subprogram part following a ``CONTAINS`` statement
        declaring member procedures
    prefix : iterable, optional
        Prefix specifications for the procedure
    bind : optional
        Bind information (e.g., for Fortran ``BIND(C)`` annotation).
    result_name : str, optional
        The name of the result variable for functions.
    is_function : bool, optional
        Flag to indicate this is a function instead of subroutine
        (in the Fortran sense). Defaults to `False`.
    ast : optional
        Frontend node for this subroutine (from parse tree of the frontend).
    source : :any:`Source`
        Source object representing the raw source string information from the
        read file.
    parent : :any:`Scope`, optional
        The enclosing parent scope of the subroutine, typically a :any:`Module`
        or :any:`Subroutine` object. Declarations from the parent scope remain
        valid within the subroutine's scope (unless shadowed by local
        declarations).
    rescope_symbols : bool, optional
        Ensure that the type information for all :any:`TypedSymbol` in the
        subroutine's IR exist in the subroutine's scope or the scope's parents.
        Defaults to `False`.
    symbol_attrs : :any:`SymbolTable`, optional
        Use the provided :any:`SymbolTable` object instead of creating a new
    incomplete : bool, optional
        Mark the object as incomplete, i.e. only partially parsed. This is
        typically the case when it was instantiated using the :any:`Frontend.REGEX`
        frontend and a full parse using one of the other frontends is pending.
    parser_classes : :any:`RegexParserClass`, optional
        Provide the list of parser classes used during incomplete regex parsing
    """

    def __init__(
            self, name, args=None, docstring=None, spec=None, body=None,
            contains=None, prefix=None, bind=None, result_name=None,
            is_function=False, ast=None, source=None, parent=None,
            symbol_attrs=None, rescope_symbols=False, incomplete=False, parser_classes=None
    ):
        super().__init__(parent=parent)

        if symbol_attrs:
            self.symbol_attrs.update(symbol_attrs)

        self.__initialize__(
            name=name, args=args, docstring=docstring, spec=spec, body=body,
            contains=contains,  prefix=prefix, bind=bind, result_name=result_name,
            is_function=is_function, ast=ast, source=source,
            rescope_symbols=rescope_symbols, incomplete=incomplete, parser_classes=parser_classes
        )

    def __initialize__(
            self, name, docstring=None, spec=None, contains=None,
            ast=None, source=None, rescope_symbols=False, incomplete=False, parser_classes=None,
            body=None, args=None, prefix=None, bind=None, result_name=None, is_function=False
    ):
        # First, store additional Subroutine-specific properties
        self._dummies = as_tuple(a.lower() for a in as_tuple(args))  # Order of dummy arguments
        self.prefix = as_tuple(prefix)
        self.bind = bind
        self.result_name = result_name
        self.is_function = is_function

        # Make sure 'result_name' is defined if it's a function
        if self.result_name is None and self.is_function:
            self.result_name = name

        # Additional IR components
        if body is not None and not isinstance(body, ir.Section):
            body = ir.Section(body=body)
        self.body = body

        super().__initialize__(
            name=name, docstring=docstring, spec=spec, contains=contains,
            ast=ast, source=source, rescope_symbols=rescope_symbols,
            incomplete=incomplete, parser_classes=parser_classes
        )

    def __getstate__(self):
        _ignore = ('_ast', '_parent')
        return dict((k, v) for k, v in self.__dict__.items() if k not in _ignore)

    def __setstate__(self, s):
        self.__dict__.update(s)

        self._ast = None

        # Re-register all encapulated member procedures
        for member in self.members:
            self.symbol_attrs[member.name] = SymbolAttributes(ProcedureType(procedure=member))

        # Ensure that we are attaching all symbols to the newly create ``self``.
        self.rescope_symbols()


    def _infer_allocatable_shapes(self):
        """
        Infer variable symbol shapes from allocations of ``allocatable`` arrays.
        """
        for alloc in FindNodes(ir.Allocation).visit(self.body):
            for v in alloc.variables:
                if isinstance(v, sym.Array):
                    if alloc.data_source:
                        new_shape = alloc.data_source.type.shape
                    else:
                        new_shape = v.dimensions

                    # Update the type to inject shape info into symbol table
                    v.type = v.type.clone(shape=new_shape)

    @classmethod
    def from_omni(cls, ast, raw_source, definitions=None, parent=None, type_map=None):
        """
        Create :any:`Subroutine` from :any:`OMNI` parse tree

        Parameters
        ----------
        ast :
            The OMNI parse tree
        raw_source : str
            Fortran source string
        definitions : list
            List of external :any:`Module` to provide derived-type and procedure declarations
        parent : :any:`Scope`, optional
            The enclosing parent scope of the subroutine, typically a :any:`Module`.
        type_map : dict, optional
            A mapping from type hash identifiers to type definitions, as provided in
            OMNI's ``typeTable`` parse tree node
        """
        type_map = type_map or {}
        if ast.tag != 'FfunctionDefinition':
            ast = ast.find('globalDeclarations/FfunctionDefinition')
        return parse_omni_ast(
            ast=ast, definitions=definitions, raw_source=raw_source,
            type_map=type_map, scope=parent
        )

    @classmethod
    def from_fparser(cls, ast, raw_source, definitions=None, pp_info=None, parent=None):
        """
        Create :any:`Subroutine` from :any:`FP` parse tree

        Parameters
        ----------
        ast :
            The FParser parse tree
        raw_source : str
            Fortran source string
        definitions : list
            List of external :any:`Module` to provide derived-type and procedure declarations
        pp_info :
            Preprocessing info as obtained by :any:`sanitize_input`
        parent : :any:`Scope`, optional
            The enclosing parent scope of the subroutine, typically a :any:`Module`.
        """
        if ast.__class__.__name__ not in ('Subroutine_Subprogram', 'Function_Subprogram'):
            ast = get_fparser_node(ast, ('Subroutine_Subprogram', 'Function_Subprogram'))
        # Note that our Fparser interface returns a tuple with the
        # Subroutine object always last but potentially containing
        # comments before the Subroutine object
        return parse_fparser_ast(
            ast, pp_info=pp_info, definitions=definitions,
            raw_source=raw_source, scope=parent
        )[-1]

    @classmethod
    def from_regex(cls, raw_source, parser_classes=None, parent=None):
        """
        Create :any:`Subroutine` from source regex'ing

        Parameters
        ----------
        raw_source : str
            Fortran source string
        parent : :any:`Scope`, optional
            The enclosing parent scope of the subroutine, typically a :any:`Module`.
        """
        ir_ = parse_regex_source(raw_source, parser_classes=parser_classes, scope=parent)
        return [node for node in ir_.body if isinstance(node, cls)][0]

    def register_in_parent_scope(self):
        """
        Insert the type information for this object in the parent's symbol table

        If :attr:`parent` is `None`, this does nothing.
        """
        if self.parent:
            self.parent.symbol_attrs[self.name] = SymbolAttributes(self.procedure_type)

    def clone(self, **kwargs):
        """
        Create a copy of the subroutine with the option to override individual
        parameters.

        Parameters
        ----------
        **kwargs :
            Any parameters from the constructor of :any:`Subroutine`.

        Returns
        -------
        :any:`Subroutine`
            The cloned subroutine object.
        """
        # Collect all properties bespoke to Subroutine
        if self.argnames and 'args' not in kwargs:
            kwargs['args'] = self.argnames
        if self.body and 'body' not in kwargs:
            kwargs['body'] = self.body
        if self.prefix and 'prefix' not in kwargs:
            kwargs['prefix'] = self.prefix
        if self.bind and 'bind' not in kwargs:
            kwargs['bind'] = self.bind
        if self.result_name and 'result_name' not in kwargs:
            kwargs['result_name'] = self.result_name
        if self.is_function and 'is_function' not in kwargs:
            kwargs['is_function'] = self.is_function

        # Rebuild body (other IR components are taken care of in super class)
        if 'body' in kwargs:
            kwargs['body'] = Transformer({}, rebuild_scopes=True).visit(kwargs['body'])

        # Escalate to parent class
        return super().clone(**kwargs)

    @property
    def _canonical(self):
        """
        Base definition for comparing :any:`Subroutine` objects.
        """
        return (
            self.name, self.is_function, self._dummies, self.prefix,
            self.bind, self.docstring, self.spec, self.body,
            self.contains, self.symbol_attrs
        )

    def __eq__(self, other):
        if isinstance(other, Subroutine):
            return self._canonical == other._canonical
        return super().__eq__(other)

    def __hash__(self):
        return hash(self._canonical)

    @property
    def procedure_symbol(self):
        """
        Return the procedure symbol for this subroutine
        """
        return sym.Variable(name=self.name, type=SymbolAttributes(self.procedure_type), scope=self.parent)

    @property
    def procedure_type(self):
        """
        Return the :any:`ProcedureType` of this subroutine
        """
        return ProcedureType(procedure=self)

    @property
    def return_type(self):
        """
        Return the return_type of this subroutine
        """
        if not self.is_function:
            return None
        return self.symbol_attrs.get(self.result_name)

    variables = ProgramUnit.variables

    @variables.setter
    def variables(self, variables):
        """
        Set the variables property and ensure that the internal declarations match.

        Note that arguments also count as variables and therefore any
        removal from this list will also remove arguments from the subroutine signature.
        """
        # Use the parent's property setter
        super(self.__class__, self.__class__).variables.__set__(self, variables)

        # Filter the dummy list in case we removed an argument
        varnames = [str(v.name).lower() for v in variables]
        self._dummies = as_tuple(arg for arg in self._dummies if str(arg).lower() in varnames)

    @property
    def arguments(self):
        """
        Return arguments in order of the defined signature (dummy list).
        """

        #Load symbol_map
        #Note that if the map is not loaded, Python will recreate it for every arguement,
        #resulting in a large overhead.
        symbol_map = self.symbol_map
        return as_tuple(symbol_map.get(arg, sym.Variable(name=arg)) for arg in self._dummies)

    @arguments.setter
    def arguments(self, arguments):
        """
        Set the arguments property and ensure that internal declarations and signature match.

        Note that removing arguments from this property does not actually remove declarations.
        """
        # FIXME: This will fail if one of the argument is declared via an interface!

        # First map variables to existing declarations
        declarations = FindNodes((ir.VariableDeclaration, ir.ProcedureDeclaration)).visit(self.spec)
        decl_map = dict((v, decl) for decl in declarations for v in decl.symbols)

        arguments = as_tuple(arguments)
        for arg in arguments:
            if arg not in decl_map:
                # By default, append new variables to the end of the spec
                assert arg.type.intent is not None
                if isinstance(arg.type, ProcedureType):
                    new_decl = ir.ProcedureDeclaration(symbols=(arg, ))
                else:
                    new_decl = ir.VariableDeclaration(symbols=(arg, ))
                self.spec.append(new_decl)

        # Set new dummy list according to input
        self._dummies = as_tuple(arg.name.lower() for arg in arguments)

    @property
    def argnames(self):
        """
        Return names of arguments in order of the defined signature (dummy list)
        """
        return [a.name for a in self.arguments]

    members = ProgramUnit.subroutines

    @property
    def ir(self):
        """
        All components of the intermediate representation in this subroutine
        """
        return (self.docstring, self.spec, self.body, self.contains)

    @property
    def interface(self):
        """
        Interface object that defines the `Subroutine` signature in header files.
        """

        # Remove all local variable declarations from interface routine spec
        # and duplicate all argument symbols within a new subroutine scope
        arg_names = [arg.name for arg in self.arguments]
        routine = Subroutine(name=self.name, args=arg_names, spec=None, body=None)
        decl_map = {}
        for decl in FindNodes((ir.VariableDeclaration, ir.ProcedureDeclaration)).visit(self.spec):
            if any(v.name in arg_names for v in decl.symbols):
                assert all(v.name in arg_names and v.type.intent is not None for v in decl.symbols), \
                    "Declarations must have intents and dummy and local arguments cannot be mixed."
                # Replicate declaration with re-scoped variables
                variables = as_tuple(v.clone(scope=routine) for v in decl.symbols)
                decl_map[decl] = decl.clone(symbols=variables)
            else:
                decl_map[decl] = None  # Remove local variable declarations
        routine.spec = Transformer(decl_map).visit(self.spec)
        return ir.Interface(body=(routine,))

    def enrich(self, definitions, recurse=False):
        """
        Apply :any:`ProgramUnit.enrich` and expand enrichment to calls declared
        via interfaces

        Parameters
        ----------
        definitions : list of :any:`ProgramUnit`
            A list of all available definitions
        recurse : bool, optional
            Enrich contained scopes
        """
        # First, enrich imported symbols
        super().enrich(definitions, recurse=recurse)

        # Secondly, take care of procedures that are declared via interface block includes
        # and therefore are not discovered via module imports
        definitions_map = CaseInsensitiveDict((r.name, r) for r in as_tuple(definitions))
        with pragmas_attached(self, ir.CallStatement, attach_pragma_post=False):
            for call in FindNodes(ir.CallStatement).visit(self.body):

                # Clone symbol to ensure Deferred symbols are
                # recognised ProcedureSymbols
                symbol = call.name.clone()
                routine = definitions_map.get(symbol.name)

                if not routine and symbol.parent:
                    # Type-bound procedure: try to obtain procedure from typedef
                    if (dtype := symbol.parent.type.dtype) is not BasicType.DEFERRED:
                        if (typedef := dtype.typedef) is not BasicType.DEFERRED:
                            if proc_symbol := typedef.variable_map.get(symbol.name_parts[-1]):
                                if (dtype := proc_symbol.type.dtype) is not BasicType.DEFERRED:
                                    if dtype.procedure is not BasicType.DEFERRED:
                                        routine = dtype.procedure

                is_not_enriched = (
                    symbol.scope is None or                         # No scope attached
                    symbol.type.dtype is BasicType.DEFERRED or      # Wrong datatype
                    symbol.type.dtype.procedure is not routine      # ProcedureType not linked
                )

                # Always update the call symbol to ensure it is up-to-date
                call._update(name=symbol)

                # Skip already enriched symbols and routines without definitions
                if not (routine and is_not_enriched):
                    debug('Cannot enrich call to %s', symbol)
                    continue

                # Remove existing symbol from symbol table if defined in interface block
                for node in [node for intf in self.interfaces for node in intf.body]:
                    if getattr(node, 'name', None) == symbol:
                        if node.parent == self:
                            node.parent = None

                # Need to update the call's symbol to establish link to routine
                symbol = symbol.clone(scope=self, type=symbol.type.clone(dtype=routine.procedure_type))
                call._update(name=symbol)

        # Rebuild local symbols to ensure correct symbol types
        self.body = ExpressionTransformer(inplace=True).visit(self.body)

    def __repr__(self):
        """
        String representation.
        """
        return f'{"Function" if self.is_function else "Subroutine"}:: {self.name}'
